@@ -6,10 +6,11 @@ Released under the GNU LGPL, version 3 or later (see COPYING).
 """
 from __future__ import absolute_import
 
-import json
-import logging
+import json, logging, copy
 from datetime import datetime, timedelta, tzinfo
 
+from onctuous.schema import Schema
+
 import boto
 from boto.dynamodb.item import Item
 from boto.exception import DynamoDBResponseError
@@ -82,9 +83,9 @@ class autoincrement_int(int):
     Inserting a new item issues an atomic add on the '__max_hash_key__' value.
     Its new value is returned and used as the primary key for the new elem.
 
-    Note that hash_key_value is set to '-1' while ``__max_hash_key__`` initial
-    value is 0. This will element at key '0' unused. It's actually a garbage item
-    for cases where a value is manually added to an unitialized index.
+    Note that hash_key_value is set to '-1' while ``__max_hash_key__`` implicit
+    initial value is 0. This guarantees element at key '0' is unused. It's actually
+    a garbage item for cases where a value is manually added to an unitialized index.
     """
 
 _JSON_TYPES = frozenset([list, dict])
@@ -104,8 +105,7 @@ class UTC(tzinfo):
 
 utc_tz = UTC()
 
-
-def _get_proto_value(schema_type):
+def _get_proto_value(schema_entry):
     """Return a prototype value matching what schema_type will be serialized
     as in DynamoDB:
 
@@ -114,51 +114,18 @@ def _get_proto_value(schema_type):
         datetime), an empty string (this is what they're stored as in the DB).
     """
     # Those types must be serialized as strings
-    if schema_type in _JSON_TYPES:
+    if schema_entry in _JSON_TYPES or type(schema_entry) in _JSON_TYPES:
         return u""
 
-    if schema_type is datetime:
+    if schema_entry is datetime:
         return u""
 
     # Regular string/number
-    return schema_type()
-
-
-def _get_default_value(schema_type, default=None, name=''):
-    """Return a default value matching schema_type or default if provided:
-
-      - For datetime.datetime, it's NOW.
-      - For container types, it's an empty container.
-      - For strings, it's an empty string.
-      - For numbers, it's zero.
-
-    This function may raise TypeError exception if:
-
-       - default was callable and required arguments
-       - default or its return value is not an instance of schema_type
-
-    :param schema_type: class object to instanciate
-    :param default: default value. May be a value or a callable (functions, class, ...) It must *NOT* require an any argument and it's type must match schema_type
-    :param name: field name to use in exception string if applicable
-
-    """
-    if default is not None:
-        # If default is callable(function, constructor, ...), try to dereference it
-        if callable(default):
-            # Might raise a "TypeError" if arguments were needed
-            default = default()
-        # Check default value consitency
-        if not isinstance(default, schema_type):
-            raise TypeError("Expected default value of type {}, got value: '{}' of type: {} for field: '{}'".format(schema_type, str(default), type(default), name))
-        else:
-            return default
-
-    if schema_type is datetime:
-        # Current Timestamp
-        return datetime.now(tz=utc_tz)
-
-    return schema_type()
+    if type(schema_entry) is type:
+        return schema_entry()
 
+    # callable validator
+    return _get_proto_value((type(schema_entry(None))))
 
 def _python_to_dynamodb(value):
     """Convert a Python object to a representation suitable to direct storage
@@ -203,45 +170,68 @@ def _python_to_dynamodb(value):
     return None
 
 
-def _dynamodb_to_python(schema_type, value):
+def _dynamodb_to_python(schema_entry, value):
     """Convert a DynamoDB attribute value to a Python object, according to a
     type from a DynamoDBModel schema.
 
-    If value is None (usually because the attribute was missing in the first
-    place), a default value is returned (empty string or set, or 0, depending
-    on what the type is).
+    If ``schema_entry`` is a type (``int`` for example) a datetime or a list/dict
+    validator, ``value`` is automatically de-serialized.
 
-    ``_dynamodb_to_python(t, _python_to_dynamodb(v)) == v`` for any v.
-
-    :param schema_type: A type supported by the mapper
-
-    .. (TODO Clearly list those).
+    The resulting value is then passed to the validation engine. If needed, make
+    sure that your validator starts by manually coerce the input value.
 
+    :param schema_entry: A type or validator supported by the mapper
     :param value: The DynamoDB attribute to convert to a Python object.
         May be ``None``.
 
-    :return: An instance of ``schema_type``.
+    :return: coerced and validated ``value``.
     """
-    # This means a missing attribute for a _JSON_TYPES object is valid,
-    # and results in an empty sequence. Is that a bad thing?
-    if value is None:
-        return None
+    schema_type = type(schema_entry)
 
+    # Handle json related type
     if schema_type in _JSON_TYPES:
-        return schema_type(json.loads(value))
+        # looks like a validator => load and validate it
+        if value is None:
+            value = schema_type()
+        else:
+            value = schema_type(json.loads(value))
+        return Schema(schema_entry)(value)
+    elif schema_entry in _JSON_TYPES:
+        # basic type => just load it and return
+        if value is None:
+            return schema_entry()
+        return schema_entry(json.loads(value))
 
-    if schema_type is datetime:
-        # Parse TZ-aware isoformat
+    # Handle this f** datetime sh** (sorry for expressing my thougts aloud)
+    # if we enter this, it means this is not a validator so return directly
+    if schema_entry is datetime:
+        if value is None:
+            return datetime.now(tz=utc_tz)
+        else:
+            # Parse TZ-aware isoformat
 
-        # strptime doesn't support timezone parsing (%z flag), so we're forcing
-        # the strings in the database to be UTC (+00:00) for now.
-        # TODO Handle arbitrary timezones (with manual parsing).
-        if value.endswith('Z'):
-            value = value[:-1] + '+00:00'
-        return datetime.strptime(
-            value, "%Y-%m-%dT%H:%M:%S.%f+00:00").replace(tzinfo=utc_tz)
+            # strptime doesn't support timezone parsing (%z flag), so we're forcing
+            # the strings in the database to be UTC (+00:00) for now.
+            # TODO Handle arbitrary timezones (with manual parsing).
+            if value.endswith('Z'):
+                value = value[:-1] + '+00:00'
+            return datetime.strptime(
+                value, "%Y-%m-%dT%H:%M:%S.%f+00:00").replace(tzinfo=utc_tz)
+
+    # handle "base" type like int and unicode: if no value, load "neutral" value
+    # for basic type, if coercion is OK, validation is done. Hence the "return" shortcut
+    # do it only once json and dates are done as json is only a special case of this one
+    if schema_type is type:
+        if value is None:
+            return schema_entry()
+        return schema_entry(value)
+
+    # if this is a regular callable, run it on the input as a validator
+    elif callable(schema_entry):
+        return schema_entry(value)
+
+    raise SchemaError("Invalid schema entry {}; can not load value {}".format(schema_entry, value))
 
-    return schema_type(value)
 
 
 class ConnectionBorg(object):
@@ -262,6 +252,7 @@ class ConnectionBorg(object):
         "_aws_secret_access_key": None,
         "_region": None,
         "_connection": None,
+        "_tables_cache": {},
     }
 
     def __init__(self):
@@ -377,7 +368,9 @@ class ConnectionBorg(object):
 
     def get_table(self, name):
         """Return the table with the requested name."""
-        return self._get_connection().get_table(name)
+        if name not in self._tables_cache:
+            self._tables_cache[name] = self._get_connection().get_table(name)
+        return self._tables_cache[name]
 
 
 class DynamoDBModel(object):
@@ -444,11 +437,6 @@ class DynamoDBModel(object):
         {
             "id": "e1m2"
         }
-
-
-    .. TODO Add checks for common error cases:
-        - Wrong datatypes in the schema
-        - hash_key/range_key incorrectly defined
     """
 
     # TODO Add checks to the various methods so that meaningful error messages
@@ -462,11 +450,11 @@ class DynamoDBModel(object):
 
     def __init__(self, **kwargs):
         """Create an instance of the model. All fields defined in the schema
-        are created. By order of prioritym its value will be loaded from:
+        are created. By order of priority its value will be loaded from:
 
             - kwargs
             - __defaults__
-            - mapper's default (0, empty string, empty set, ...)
+            - None. Warning, this will most likely cause failures at save time
 
         We're supplying this method to avoid the need for extra checks in save and
         ease object initial creation.
@@ -482,51 +470,33 @@ class DynamoDBModel(object):
 
         for (name, type_) in schema.iteritems():
             if name in kwargs:
-                default = kwargs.get(name)
+                value = kwargs.get(name)
+            elif name in defaults:
+                template = defaults[name]
+                if callable(template):
+                    value = template()
+                else:
+                    value = copy.deepcopy(template)
             else:
-                default = defaults[name] if name in defaults else None
-            value = _dynamodb_to_python(type_, default) # just in case, ensure types. help migration from 'from_dict'
-            value = _get_default_value(type_, value, name=name)
+                value = None
             setattr(self, name, value)
 
         # instanciate the migrator only once per model *after* initialization
+        # as it assumes a fully initialized model
         if isinstance(cls.__migrator__, type):
             cls.__migrator__ = cls.__migrator__(cls)
 
-    @classmethod
-    def _from_db_dict(cls, raw_data):
-        """Build an instance from a dict-like mapping, according to the class's
-        schema. Objects created with this method are considered as comming from
-        the DB. The initial state is persisted in ``self._raw_data``.
-        If a ``__migrator__`` has been declared, migration is triggered on a copy
-        of the raw data.
-
-        Default values are used for anything that's missing from the dict
-        (see DynamoDBModel class docstring).
-
-        Direct use of this method should be avoided as much as possible but still
-        may be usefull for "deep copy".
-
-        Overload this method if you need a special (de-)serialization semantic
-
-        :param raw_data: Raw db dict
+    def validate(self):
+        """Return a ``dict`` of validated fields if validators passes. Otherwise
+        ``InvalidList`` is raised.
         """
-        #FIXME: type check. moving to __init__ syntax may break some implementations
-        instance = cls()
-        instance._raw_data = raw_data
-
-        #If a migrator is registered, trigger it
-        if cls.__migrator__ is not None:
-           raw_data = cls.__migrator__(raw_data)
-
-        # de-serialize data
-        for (name, type_) in cls.__schema__.iteritems():
-            # Set the value if we got one from DynamoDB. Otherwise, stick with the default
-            value = _dynamodb_to_python(type_, raw_data.get(name)) # de-serialize
-            value = _get_default_value(type_, value, name=name)    # match value
-            setattr(instance, name, value)
-
-        return instance
+        # load schema
+        schema = self.__schema__
+        validate = Schema(schema)
+        # load schema data from self
+        data = {str(key):getattr(self, str(key)) for key in schema}
+        # return validated data (or raise)
+        return validate(data)
 
     @classmethod
     def get(cls, hash_key_value, range_key_value=None, consistent_read=False):
@@ -693,6 +663,40 @@ class DynamoDBModel(object):
             if d[hash_key_name] != MAGIC_KEY or cls.__schema__[hash_key_name] != autoincrement_int
         )
 
+    @classmethod
+    def _from_db_dict(cls, raw_data):
+        """Build an instance from a dict-like mapping, according to the class's
+        schema. Objects created with this method are considered as comming from
+        the DB. The initial state is persisted in ``self._raw_data``.
+        If a ``__migrator__`` has been declared, migration is triggered on a copy
+        of the raw data.
+
+        Default values are used for anything that's missing from the dict
+        (see DynamoDBModel class docstring).
+
+        Direct use of this method should be avoided as much as possible but still
+        may be usefull for "deep copy".
+
+        Overload this method if you need a special (de-)serialization semantic
+
+        :param raw_data: Raw db dict
+        """
+        #FIXME: type check. moving to __init__ syntax may break some implementations
+        instance = cls()
+        instance._raw_data = raw_data
+
+        #If a migrator is registered, trigger it
+        if cls.__migrator__ is not None:
+           raw_data = cls.__migrator__(raw_data)
+
+        # de-serialize data
+        for (name, type_) in cls.__schema__.iteritems():
+            # Set the value if we got one from DynamoDB. Otherwise, stick with the default
+            value = _dynamodb_to_python(type_, raw_data.get(name)) # de-serialize
+            setattr(instance, name, value)
+
+        return instance
+
     def _to_db_dict(self):
         """Return a dict representation of the object according to the class's
         schema, suitable for direct storage in DynamoDB.
@@ -702,14 +706,8 @@ class DynamoDBModel(object):
 
         Overload this method if you need a special serialization semantic
         """
-        out = {}
-        for name in type(self).__schema__:
-            value = _python_to_dynamodb(getattr(self, name))
-            if value is not None:
-                # DynamoDB can't store empty sets/strings, so we must
-                # represent them as missing attributes
-                out[name] = value
-        return out
+        data = self.validate()
+        return {key: _python_to_dynamodb(val) for key, val in data.iteritems() if val or val == 0}
 
     def to_json_dict(self):
         """Return a dict representation of the object, suitable for JSON
@@ -734,7 +732,7 @@ class DynamoDBModel(object):
                 out[name] = value
         return out
 
-    def _save_autoincrement_hash_key(self, item):
+    def _save_autoincrement_hash_key(self):
         """Compute an autoincremented hash_key for an item and save it to the DB.
 
         To achieve this goal, we keep a special object at ``hash_key=MAGIC_KEY``
@@ -745,31 +743,30 @@ class DynamoDBModel(object):
         The new value is send back to us and used as the hash_key for elem
         """
         counter_key = '__max_hash_key__'
+        hk_name = self.__hash_key__
+        table = ConnectionBorg().get_table(self.__table__)
         tries = 0
 
         while tries < MAX_RETRIES:
             tries += 1
-            # Create a 'new item' with key=0 and trigger an atomic increment
+            # Create a 'new item' with key=-1 and trigger an atomic increment
             # This spares one read unit :)
-            max_hash_item = item.table.new_item(MAGIC_KEY)
+            max_hash_item = Item(MAGIC_KEY)
             max_hash_item.add_attribute(counter_key, 1)
             max_hash_item = max_hash_item.save(return_values='ALL_NEW')
             # We just reserved that value for the hash key
             hash_key = max_hash_item['Attributes'][counter_key]
-            item[item.hash_key_name] = hash_key
+            setattr(self, hk_name, autoincrement_int(hash_key))
 
             try:
                 # Make sure this primary key was not 'stolen' by a direct DB access
-                item.put({item.hash_key_name: False})
-                dblog.debug("Saved autoinc (%s) in table %s", hash_key, item.table)
+                self.save(raise_on_conflict=True)
+                dblog.debug("Saved autoinc (%s) in table %s", hash_key, table.name)
                 return
-            except DynamoDBResponseError as e:
-                if e.error_code != "ConditionalCheckFailedException":
-                    # Unhandled exception
-                    raise
+            except ConflictError as e:
                 log.debug(
                     "table=%s, An item seems to have been manually inserted at index %s (%s).",
-                    item.table, item[item.hash_key_name], tries)
+                    table.name, hash_key, tries)
 
         # This table auto-incr has been screwed up...
         raise MaxRetriesExceededError()
@@ -807,48 +804,48 @@ class DynamoDBModel(object):
         hash_key = cls.__hash_key__
         range_key = cls.__range_key__
         table = ConnectionBorg().get_table(cls.__table__)
+
+        # Detect magic elem manual overwrite
+        if schema[hash_key] == autoincrement_int and getattr(self, hash_key) == MAGIC_KEY:
+            raise SchemaError("Index {} is reserved in table with autoincrementing key".format(MAGIC_KEY))
+        # We're inserting a new item in an autoincrementing table.
+        if schema[hash_key] == autoincrement_int and getattr(self, hash_key) is None:
+            # allocate the index and recursively call this method
+            return self._save_autoincrement_hash_key()
+
+
         item_data = self._to_db_dict()
         item = Item(table, attrs=item_data)
 
-        # Detect magic elem manual overwrite
-        if schema[hash_key] == autoincrement_int and item_data[hash_key] == MAGIC_KEY:
-            raise SchemaError()#FIXME: probably not the best exception to throw
-        # We're inserting a new item in an autoincrementing table.
-        if schema[hash_key] == autoincrement_int and item_data[hash_key] == 0:
-            # Compute the index and save the object
-            self._save_autoincrement_hash_key(item)
-            # Update the primary key so that it reflects what it was saved as.
-            setattr(self, hash_key, item[hash_key])
         # Regular save
-        else:
-            if raise_on_conflict:
-                if self._raw_data:
-                    expected_values = self._raw_data
-                    # Empty strings/sets must be represented as missing values
-                    for name in schema.iterkeys():
-                        if name not in expected_values:
-                            expected_values[name] = False
-                else:
-                    # Forbid overwrites: do a conditional write on
-                    # "this hash_key doesn't exist"
-                    allow_overwrite = False
-                    expected_values = {hash_key: False}
-                    if range_key:
-                        expected_values[range_key] = False
-            try:
-                item.put(expected_values)
-            except DynamoDBResponseError as e:
-                if e.error_code == "ConditionalCheckFailedException":
-                    if allow_overwrite:
-                        # Conflict detected
-                        raise ConflictError(item)
-                    # Forbidden overwrite
-                    raise OverwriteError(item)
-                # Unhandled exception
-                raise
+        if raise_on_conflict:
+            if self._raw_data:
+                expected_values = self._raw_data
+                # Empty strings/sets must be represented as missing values
+                for name in schema.iterkeys():
+                    if name not in expected_values:
+                        expected_values[name] = False
+            else:
+                # Forbid overwrites: do a conditional write on
+                # "this hash_key doesn't exist"
+                allow_overwrite = False
+                expected_values = {hash_key: False}
+                if range_key:
+                    expected_values[range_key] = False
+        try:
+            item.put(expected_values)
+        except DynamoDBResponseError as e:
+            if e.error_code == "ConditionalCheckFailedException":
+                if allow_overwrite:
+                    # Conflict detected
+                    raise ConflictError(item)
+                # Forbidden overwrite
+                raise OverwriteError(item)
+            # Unhandled exception
+            raise
 
         # Update Raw_data to reflect DB state on success
-        self._raw_data = self._to_db_dict()
+        self._raw_data = item_data
 
         hash_key_value = getattr(self, hash_key)
         range_key_value = getattr(self, range_key, None) if range_key else None
