@@ -172,6 +172,13 @@ class SchemaValidators(DynamoDBModel):
         "scores": [InRange(min=0, max=100)],
     }
 
+class SchemaValidatorsCoerceLong(DynamoDBModel):
+    __table__ = "SchemaValidatorsCoerceLong"
+    __hash_key__ = "id"
+    __schema__ = {
+        "id": Coerce(long),
+    }
+
 class SchemaValidatorsBad(DynamoDBModel):
     __table__ = "SchemaValidatorsBad"
     __hash_key__ = "name"
@@ -504,6 +511,28 @@ class TestConnectionBorg(unittest.TestCase):
 
         table.refresh.assert_called_once_with(wait_for_active=True)
 
+    @mock.patch("dynamodb_mapper.model.boto")
+    def test_create_table_validators_coerce_long(self, m_boto):
+        # Issue #20.
+        m_connection = m_boto.connect_dynamodb.return_value
+        m_create_schema = m_connection.create_schema
+        m_create_table = m_connection.create_table
+
+        table = ConnectionBorg().create_table(
+            SchemaValidatorsCoerceLong, 10, 5, True)
+
+        m_create_schema.assert_called_once_with(
+            hash_key_name=SchemaValidatorsCoerceLong.__hash_key__,
+            hash_key_proto_value=0L,
+            range_key_name=None,
+            range_key_proto_value=None,
+        )
+
+        m_create_table.assert_called_once_with(
+            SchemaValidatorsCoerceLong.__table__, m_create_schema.return_value, 10, 5)
+
+        table.refresh.assert_called_once_with(wait_for_active=True)
+
 
 class TestTypeConversions(unittest.TestCase):
     def test_python_to_dynamodb_number(self):
